package tsdb;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.LocalDateTime;
import java.util.UUID;

public class TsdbApiEntry {
    private static final Logger LOG = LoggerFactory.getLogger(TsdbApiEntry.class);

    private String metricName = "";
    private String dataDir = "data";
    private long partSize = 10000;
    private String correlationId = "";

    public String getMetricName() {
        return metricName;
    }

    public void setMetricName(String metricName) {
        LOG.debug("Setting metricName to {}, correlation ID = {}", metricName, correlationId);
        this.metricName = metricName;
    }

    public String getDataDir() {
        return dataDir;
    }

    public void setDataDir(String dataDir) {
        LOG.debug("Setting dataDir to {}, correlation ID = {}", dataDir, correlationId);
        if (StringUtils.isEmpty(dataDir)) {
            throw new IllegalArgumentException("dataDir must not be empty");
        }
        this.dataDir = dataDir;
    }

    public long getPartSize() {
        return partSize;
    }

    public void setPartSize(long partSize) {
        LOG.debug("Setting partSize to {}, correlation ID = {}", partSize, correlationId);
        if (partSize <= 0) {
            throw new IllegalArgumentException("partSize must be greater than 0");
        }
        this.partSize = partSize;
    }

    String getCorrelationId() {
        return correlationId;
    }

    public TsdbApiEntry() {
        correlationId = UUID.randomUUID().toString();
        LOG.debug("Created TsdbApiEntry, correlation ID = {}", correlationId);
    }

    @Override
    public String toString() {
        return "TsdbApiEntry{" +
                "metricName=" + Util.escapeIntoSingleQuotedString(metricName) +
                ", dataDir=" + Util.escapeIntoSingleQuotedString(dataDir) +
                ", partSize=" + partSize +
                ", correlationId=" + Util.escapeIntoSingleQuotedString(correlationId) +
                '}';
    }

    /// Creates a new stream for writing points data.
    /// @param name The name of the point.
    /// @param timestampOffset The timestamp offset, in milliseconds.
    /// @param interval The interval between points, in nanoseconds.
    /// @return The new stream.
    public TsdbApiEntryStream newStream(String name, long timestampOffset, long interval) {
        try {
            String dirName = String.format("%s-%s", name, Util.getFormattedTimestamp(LocalDateTime.now()));
            String streamBasePath = Util.pathCombine(dataDir, dirName);
            return new TsdbApiEntryStream(this, streamBasePath, name, timestampOffset, interval);
        } catch (Exception e) {
            LOG.error("Failed to create new stream for name = {}, timestampOffset = {}, interval = {}, correlation ID = {}", name, timestampOffset, interval, correlationId, e);
            throw e;
        }
    }
}
